import math as math
import random as random

from .Tools import setAttribute as setatt
from .LaplacePrior import LaplacePrior

__author__ = "Do Kester"
__year__ = 2024
__license__ = "GPL3"
__version__ = "3.2.1"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 202024 Do Kester

class ExponentialPrior( LaplacePrior ):
    """
    Exponential prior distribution.

        Pr( x ) = exp( -x / scale )

    By default scale = 1.

    The domain is [0,+Inf].
    In computational practice the domain is limited to about [0,36] scale units

    Wrapper for:
    LaplacePrior( center=0, scale=scale, limits=[0, hilim] )

    Examples
    --------
    >>> pr = ExponentialPrior()                     # scale=1.0
    >>> pr = ExponentialPrior( scale=5.0 )          # scale=5

    Attributes
    ----------
    scale : float
        scale of the exponential

    Attributes from Prior
    --------------------=
    lowLimit, highLimit, deltaP, _lowDomain, _highDomain

    Author: Do Kester.
    """

    MAXVAL = 40

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, scale=1.0, hilimit=math.inf, prior=None ):
        """
        Constructor.

        Parameters:
        scale : float
            of the exponential
        hilimit : float
            high limit
        prior : ExponentialPrior
            prior to copy (with new scale if applicable)

        """
        super( ).__init__( scale=scale, limits=[0,hilimit], prior=prior )

        self._lowDomain = 0 

    def copy( self ):
        """ Copy the prior """
        return ExponentialPrior( scale=self.scale, hilimit=self.limits[1], 
            prior=self )

    def shortName( self ) :
        return "ExponentialPrior"






import numpy as numpy
from astropy import units
from .NonLinearModel import NonLinearModel
from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.2.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2023 Do Kester


class FootballModel( NonLinearModel ):
    """
    More or less complex model for the outcome of football marches.

    The input values are a (nteams,2) list of integers. They represent 
    teams that play a match, the first at home the other away.

    For each team the complexity lists parameters

       name                    limits    default    comment
    0. Number of trials      0 < a         n/a     trials on the goal of the opponent
    1. Defensive strength    0 < b < 1      0      fraction of the trials that is stopped
    2. Midfield strength     0 < c < 2      1      relative strength of the team
    3. Home advantage        0 < d < 2      1      advantage of playing at home
    4. Strategy              0 < e < 2      1      

    Note: Computational runtime errors/warnings occur when (some of) the parameters are at 
          their limits. 

    The default values are chosen such that they dont have effect on the results.
    I.e. a model with complexity=5 and all parameters at the defaults except for 
    "trials", has the same result as a model with complexity 1 with the same "trials"
    value.

    Let p1 denote the parameters of the home team and p2 those of the away team,
    then the equations for calculating the strengths, S1 and S2, are

        S1 = a1 * sqrt( c1 * d1 / c2 ) * ( 1 - b2 ^ ( c1 * d1 / c2 ) )

        S2 = a2 * sqrt( c2 / ( c1 * d1 ) ) * ( 1 - b1 ^ ( c2 / ( c1 * d1 ) )


    Note:
    This is about the game that most of the world calls football.

    Examples
    --------
    >>> fm = FootballModel( 18 ) 
    >>> print( fm.npars )
    90

    Author : Do Kester

    Attributes
    ----------
    nteams : int
        number of teams
    complexity : int
        degree of complexity, default = 5.

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    """
    PARNAMES = ["trials", "defense", "midfield", "home", "strategy"]

    def __init__( self, nteams, complexity=5, copy=None, **kwargs ):
        """
        Calculate the score of football matches

        The number of parameters is ( nteams * complexity )

        Parameters
        ----------
        nteams : int
            number of teams
        complexity : 1 <= int <= 5
            of the model
        copy : FootballModel
            model to copy
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        if complexity < 1 or complexity > 5 :
            raise ValueError( "Complexity must be between 1 and 5, inclusive." )

        super( ).__init__( complexity * nteams, ndim=2, copy=copy, **kwargs )

        self.ndout = 2
        self.complexity = complexity
        self.nteams = nteams

    def copy( self ):
        """ Copy method.  """
        return FootballModel( self.nteams, complexity=self.complexity, copy=self )

    def getPrior( self, k ) :
        """
        Return the prior of the parameter, indicated by k modulo the complexity

        Parameters
        ---------
        k : int
            parameter number.
        """
        return self.basePrior( k % self.complexity )
        

    def __setattr__( self, name, value ):
        """
        Set attributes: degree
        """
        if name == 'complexity' :
            if value == 5 :
                setatt( self, "goals", self.goals5 )
                setatt( self, "parts", self.part5 )
            elif value == 4 :
                setatt( self, "goals", self.goals4 )
                setatt( self, "parts", self.part4 )
            elif value == 3 :
                setatt( self, "goals", self.goals3 )
                setatt( self, "parts", self.part3 )
            elif value == 2 :
                setatt( self, "goals", self.goals2 )
                setatt( self, "parts", self.part2 )
            elif value == 1 :
                setatt( self, "goals", self.goals1 )
                setatt( self, "parts", self.part1 )
            else :
                raise ValueError( "Complexity must be between 1 and 5, inclusive" )

            setatt( self, name, value, type=int )

        elif name == 'nteams' :
            setatt( self, name, value, type=int )
        elif name == 'ndout' :
            setatt( self, name, value, type=int )
        else :
            super( ).__setattr__( name, value )


    def goals1( self, xdata, par ) :
        """ attack """
        return numpy.asarray( [par[xdata[:,0]], par[xdata[:,1]]] ).transpose()

    def goals2( self, xdata, par ) :
        """ attack, defense """
        k0 = xdata[:,0] * 2
        k1 = xdata[:,1] * 2
        return numpy.asarray( [par[k0] * ( 1 - par[k1+1] ), 
                               par[k1] * ( 1 - par[k0+1] )] ).transpose()

    def goals3( self, xdata, par ) :
        """ attack, defense, midfield """
        k0 = xdata[:,0] * 3
        k1 = xdata[:,1] * 3
        ma0 = md1 = par[k0+2] / par[k1+2]
        ma1 = md0 = 1 / ma0
#        print( "3 m0  ", fmt( m0 ) )
#        print( "3 m1  ", fmt( m1 ) )

        sc0 = par[k0] * numpy.sqrt( ma0 ) * ( 1 - numpy.power( par[k1+1], md1 ) )
        sc1 = par[k1] * numpy.sqrt( ma1 ) * ( 1 - numpy.power( par[k0+1], md0 ) )
        return numpy.asarray( [sc0, sc1] ).transpose()

    def goals4( self, xdata, par ) :
        """ attack, defense, midfield, home """
        k0 = xdata[:,0] * 4
        k1 = xdata[:,1] * 4
        ma0 = md1 = ( par[k0+2] * par[k0+3] ) / par[k1+2]
        ma1 = md0 = 1 / ma0
#        print( "4 m0  ", fmt( m0 ) )
#        print( "4 m1  ", fmt( m1 ) )

        sc0 = par[k0] * numpy.sqrt( ma0 ) * ( 1 - numpy.power( par[k1+1], md1 ) )
        sc1 = par[k1] * numpy.sqrt( ma1 ) * ( 1 - numpy.power( par[k0+1], md0 ) )
        return numpy.asarray( [sc0, sc1] ).transpose()

    def goals5( self, xdata, par ) :
        """
        attack, defense, midfield, home, strategy 

        """
        k0 = xdata[:,0] * 5
        k1 = xdata[:,1] * 5

        m0 = ( par[k0+2] * par[k0+3] ) / par[k1+2]
        m1 = 1 / m0
        ma0 = m0 * par[k0+4]
        md0 = m1 * par[k0+4]
        ma1 = m1 * par[k1+4]
        md1 = m0 * par[k1+4]

#        print( "5 m0  ", fmt( m0 ), fmt( ma0 ), fmt( md0 ) )
#        print( "5 m1  ", fmt( m1 ), fmt( ma1 ), fmt( md1 ) )

        sc0 = par[k0] * numpy.sqrt( ma0 ) * ( 1 - numpy.power( par[k1+1], md1 ) )
        sc1 = par[k1] * numpy.sqrt( ma1 ) * ( 1 - numpy.power( par[k0+1], md0 ) )
        return numpy.asarray( [sc0, sc1] ).transpose()


    def baseResult( self, xdata, params ):
        """
        Returns the partials at the input value.

        The partials are the powers of x ( xdata ) from 0 to degree.

        Parameters
        ----------
        xdata : array_like [2:nteams]
            list of team ids playing against each other.
        params : array_like
            parameters for the model 

        """
        return self.goals( xdata, params )



    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        The partials are the powers of x ( xdata ) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model (ignored for LinearModels).
        parlist : array_like
            list of indices of active parameters

        to a1
        (1-b_2^((c_1*d_1)/(c_2*f_2)))*sqrt((c_1*d_1)/c_2)
        to b2
        -(a_1*c_1*d_1*sqrt((c_1*d_1)/c_2)*b_2^((c_1*d_1)/(c_2*f_2)-1))/(c_2*f_2)
        to c1
        -(a_1*d_1*(2*b_2^((d_1*c_1)/(c_2*f_2))*log(b_2)*d_1*c_1+(b_2^((d_1*c_1)/(c_2*f_2))-1)
            *c_2*f_2))/(2*c_2^2*f_2*sqrt((d_1*c_1)/c_2))
        to c2
        (a_1*c_1*d_1*((b_2^((c_1*d_1)/(f_2*c_2))-1)*f_2*c_2+2*b_2^((c_1*d_1)/(f_2*c_2))*
            log(b_2)*c_1*d_1))/(2*f_2*sqrt((c_1*d_1)/c_2)*c_2^3)
        to d1
        -(a_1*c_1*(2*b_2^((c_1*d_1)/(c_2*f_2))*log(b_2)*c_1*d_1+(b_2^((c_1*d_1)/(c_2*f_2))-1)*
            c_2*f_2))/(2*c_2^2*f_2*sqrt((c_1*d_1)/c_2))
        to f2
        (a_1*b_2^((c_1*d_1)/(c_2*f_2))*log(b_2)*c_1*d_1*sqrt((c_1*d_1)/c_2))/(c_2*f_2^2)
        """
        return self.parts( xdata, params )



    def part1( self, xdata, par ) :
        k0 = xdata[:,0]
        k1 = xdata[:,1]
        lenk = len( k0 )
        kr = range( lenk )
        phome = numpy.zeros( (lenk, self.npbase), dtype=float )
        paway = numpy.zeros( (lenk, self.npbase), dtype=float )

        phome[kr,k0] = 1.0
        paway[kr,k1] = 1.0
        return [phome,paway]    
    
    def part2( self, xdata, par ) :
        k0 = xdata[:,0] * 2
        k1 = xdata[:,1] * 2
        lenk = len( k0 )
        kr = range( lenk )
        phome = numpy.zeros( (lenk, self.npbase), dtype=float )
        paway = numpy.zeros( (lenk, self.npbase), dtype=float )

        phome[kr,k0] = 1.0 - par[k1+1]  
        paway[kr,k1] = 1.0 - par[k0+1]

        phome[kr,k1+1] = -par[k0]
        paway[kr,k0+1] = -par[k1]

        return [phome,paway]    

    def part3( self, xdata, par ) :
        k0 = xdata[:,0] * 3
        k1 = xdata[:,1] * 3
        lenk = len( k0 )
        kr = range( lenk )
        phome = numpy.zeros( (lenk, self.npbase), dtype=float )
        paway = numpy.zeros( (lenk, self.npbase), dtype=float )

        a0 = par[k0]
        a1 = par[k1]
        b0 = par[k0+1]
        b1 = par[k1+1]
        c0 = par[k0+2]
        c1 = par[k1+2]

        m0 = c0 / c1
        m1 = 1 / m0
        sm0 = numpy.sqrt( m0 )
        sm1 = numpy.sqrt( m1 )
        pp0 = numpy.power( b0, m1 )
        pp1 = numpy.power( b1, m0 )
        lb0 = numpy.log( b0 )
        lb1 = numpy.log( b1 )
        pc0 = ( pp0 - 1 ) * c0
        pc1 = ( pp1 - 1 ) * c1
        pl0 = 2 * pp0 * lb0 * c1
        pl1 = 2 * pp1 * lb1 * c0
        sc0 = 2 * sm0 * c1 * c1
        sc1 = 2 * sm1 * c0 * c0

        phome[kr,k0] += sm0 * ( 1 - pp1 )                        # df0/da0
        paway[kr,k1] += sm1 * ( 1 - pp0 )                        # df1/da1
        # df0/da1 = df1/da0 = 0 

        phome[kr,k1+1] += -a0 * sm0 * m0 * pp1 / b1              # df0/db1
        paway[kr,k0+1] += -a1 * sm1 * m1 * pp0 / b0              # df1/db0
        # df0/db0 = df1/db1 = 0 

        apps0 = a0 * ( pc1 + pl1 ) / sc0
        apps1 = a1 * ( pc0 + pl0 ) / sc1

        phome[kr,k0+2] += -apps0                                 # df0/dc0
        phome[kr,k1+2] +=  apps0 * m0                            # df0/dc1
        paway[kr,k1+2] += -apps1                                # df1/dc1
        paway[kr,k0+2] +=  apps1 * m1                            # df1/dc0

        return [phome,paway]    

    def part4( self, xdata, par ) :
        k0 = xdata[:,0] * 4
        k1 = xdata[:,1] * 4
        lenk = len( k0 )
        kr = range( lenk )
        phome = numpy.zeros( (lenk, self.npbase), dtype=float )
        paway = numpy.zeros( (lenk, self.npbase), dtype=float )

        a0 = par[k0]
        a1 = par[k1]
        b0 = par[k0+1]
        b1 = par[k1+1]
        c0 = par[k0+2]
        c1 = par[k1+2]
        d0 = par[k0+3]
#        d1 = par[k1+3]

        m0 = c0 * d0 / c1
        m1 = 1 / m0
        sm0 = numpy.sqrt( m0 )
        sm1 = numpy.sqrt( m1 )
        pp0 = numpy.power( b0, m1 )
        pp1 = numpy.power( b1, m0 )
        lb0 = numpy.log( b0 )
        lb1 = numpy.log( b1 )
        pc0 = ( pp0 - 1 ) * c0
        pc1 = ( pp1 - 1 ) * c1
        pl0 = 2 * pp0 * lb0 * c1
        pl1 = 2 * pp1 * lb1 * c0 * d0
        sc0 = 2 * sm0 * c1 * c1
        sc1 = 2 * sm1 * c0 * c0
        d02 = d0 * d0

        phome[kr,k0] = sm0 * ( 1 - pp1 )                        # df0/da0
        paway[kr,k1] = sm1 * ( 1 - pp0 )                        # df1/da1
        # df0/da1 = df1/da0 = 0 

        phome[kr,k1+1] = -a0 * sm0 * m0 * pp1 / b1              # df0/db1
        paway[kr,k0+1] = -a1 * sm1 * m1 * pp0 / b0              # df1/db0
        # df0/db0 = df1/db1 = 0 

        apps0 = a0 * ( pc1 + pl1 ) / sc0
        apps1 = a1 * ( pc0 * d0 + pl0 ) / sc1

        phome[kr,k0+2] = -apps0 * d0                            # df0/dc0
        phome[kr,k1+2] =  apps0 * m0                            # df0/dc1
        paway[kr,k1+2] = -apps1 / d02                           # df1/dc1
        paway[kr,k0+2] =  apps1 * m1 / d0                       # df1/dc0

        phome[kr,k0+3] = -apps0 * c0                             # df0/dd0
        paway[kr,k0+3] =  apps1 * c1 / ( d02 * d0 )              # df1/dd0
        # df0/dd1 = df1/dd1 = 0

        return [phome,paway]    


    def part5( self, xdata, par ) :
        """
        Derivatives copies from https://www.derivative-calculator.net

        Fh = a0*sqrt((d0*f0*c0)/c1) * (1-b1**((d0*f1*c0)/c1))
        Fa = a1*sqrt((c1*f1)/(c0*d0)) * (1-b0**((c1*f0)/(c0*d0)))

        dFh/da0 = (1-b1**((d0*f1*c0)/c1))*sqrt((d0*f0*c0)/c1)
        dFh/da1 = 0
        dFh/db0 = 0
        dFh/db1 = -(a0*c0*d0* sqrt((c0*d0*f0)/c1) *f1* b1**((c0*d0*f1)/c1-1)) / c1
        dFh/dc0 = -(a0*d0*f0*(2*b1**((d0*f1*c0)/c1)*log(b1)*d0*f1*c0+
                    (b1**((d0*f1*c0)/c1)-1)*c1))/(2*c1**2*sqrt((d0*f0*c0)/c1))
        dFh/dc1 = (a0*c0*d0*f0*((b1**((c0*d0*f1)/c1)-1)*c1+2*b1**((c0*d0*f1)/c1)
                   *log(b1)*c0*d0*f1))/(2*sqrt((c0*d0*f0)/c1)*c1**3)
        dFh/dd0 = -(a0*c0*f0*(2*b1**((c0*f1*d0)/c1)*log(b1)*c0*f1*d0+
                    (b1**((c0*f1*d0)/c1)-1)*c1))/(2*c1**2*sqrt((c0*f0*d0)/c1))
        dFh/dd1 = 0
        dFh/df0 = (a0*(1-b1**((c0*d0*f1)/c1))*c0*d0)/(2*c1*sqrt((c0*d0*f0)/c1))
        dFh/df1 = -(a0*b1**((c0*d0*f1)/c1)*log(b1)*c0*d0*sqrt((c0*d0*f0)/c1))/c1

        dFa/da0 = 0
        dFa/da1 = (1-b0**((c1*f0)/(c0*d0)))*sqrt((c1*f1)/(c0*d0))
        dFa/db0 = -(a1*c1*f0*sqrt((c1*f1)/(c0*d0))*b0**((c1*f0)/
                    (c0*d0)-1))/(c0*d0)
        dFa/db1 = 0
        dFa/dc0 = (a1*c1*f1*((b0**((c1*f0)/(d0*c0))-1)*d0*c0+2*
                    b0**((c1*f0)/(d0*c0))*log(b0)*c1*f0))/(2*d0**2*
                    sqrt((c1*f1)/(d0*c0))*c0**3)
        dFa/dc1 = -(a1*f1*(2*b0**((f0*c1)/(c0*d0))*log(b0)*f0*c1+
                    (b0**((f0*c1)/(c0*d0))-1)*c0*d0))/(2*c0**2*d0**2*
                    sqrt((f1*c1)/(c0*d0)))
        dFa/dd0 = (a1*c1*f1*((b0**((c1*f0)/(c0*d0))-1)*c0*d0+2*
                    b0**((c1*f0)/(c0*d0))*log(b0)*c1*f0))/(2*c0**2*
                    sqrt((c1*f1)/(c0*d0))*d0**3)
        dFa/dd1 = 0
        dFa/df0 = -(a1*b0**((c1*f0)/(c0*d0))*log(b0)*c1*
                    sqrt((c1*f1)/(c0*d0)))/(c0*d0)
        dFa/df1 = (a1*(1-b0**((c1*f0)/(c0*d0)))*c1)/(2*c0*d0*
                    sqrt((c1*f1)/(c0*d0)))

        """
        k0 = xdata[:,0] * 5
        k1 = xdata[:,1] * 5
        lenk = len( k0 )
        kr = range( lenk )
        phome = numpy.zeros( (lenk, self.npbase), dtype=float )
        paway = numpy.zeros( (lenk, self.npbase), dtype=float )

        a0 = par[k0]
        a1 = par[k1]
        b0 = par[k0+1]
        b1 = par[k1+1]
        c0 = par[k0+2]
        c1 = par[k1+2]
        d0 = par[k0+3]
#        d1 = par[k1+3]
        f0 = par[k0+4]
        f1 = par[k1+4]


        cd0 = c0 * d0
        ma0 = cd0 * f0 / c1
        ma1 = c1 * f1 / cd0
        sm0 = numpy.sqrt( ma0 )
        sm1 = numpy.sqrt( ma1 )
        md0 = c1 * f0 / cd0
        md1 = cd0 * f1 / c1
        pp0 = numpy.power( b0, md0 )
        pp1 = numpy.power( b1, md1 )
        lb0 = numpy.log( b0 )
        lb1 = numpy.log( b1 )
        pb0 = pp0 - 1
        pb1 = pp1 - 1

        phome[kr,k0] = -sm0 * pb1                                # df0/da0
        paway[kr,k1] = -sm1 * pb0                                # df1/da1
        # df0/da1 = df1/da0 = 0 

        phome[kr,k1+1] = -a0 * sm0 * md1 * pp1 / b1              # df0/db1
        paway[kr,k0+1] = -a1 * sm1 * md0 * pp0 / b0              # df1/db0
        # df0/db0 = df1/db1 = 0 

        phome[kr,k0+2] = -(a0*d0*f0*(2* pp1 * lb1 *cd0*f1 + pb1 *c1))/(2*c1**2* sm0 )  # df0dc0
        phome[kr,k1+2] = (a0*cd0*f0*( pb1 *c1 + 2*pp1 * lb1 *cd0*f1))/(2* sm0 *c1**3)  # df0dc1
        paway[kr,k1+2] = -(a1*f1*(2* pp0 * lb0 * f0*c1 + pb0 *cd0))/(2*cd0**2* sm1 )   # df1dc1
        paway[kr,k0+2] = (a1*c1*f1*(pb0 * cd0 + 2*pp0 * lb0 *c1*f0))/(2*cd0**2* sm1 * c0) # df1dc0

        phome[kr,k0+3] = -(a0*c0*f0*( 2* pp1 * lb1 *cd0*f1 + pb1*c1 )) /( 2*c1**2 * sm0 ) # df0dd0
        paway[kr,k0+3] = (a1*c1*f1*(pb0*cd0 + 2*pp0 * lb0 * c1*f0))/( 2*cd0**2* sm1 * d0) # df1dd0
        # df0/dd1 = df1/dd1 = 0

        phome[kr,k0+4] = -(a0* pb1 * cd0) / ( 2*c1 * sm0 )          # df0df0
        phome[kr,k1+4] = -(a0* pp1 * lb1 * cd0 * sm0 ) / c1         # df0df1
        paway[kr,k1+4] = -(a1* pb0 * c1 ) / ( 2*cd0 * sm1 )         # df1df1
        paway[kr,k0+4] = -(a1* pp0 * lb0 * c1 * sm1 ) / cd0         # df1df0

        return [phome,paway]    


    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each input (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model.

        """
        raise NotImplementedError( "Impossible. Integer values in xdata." )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return "Foolball"

    def baseParameterName( self, k ):
        """
        Return the name of the indicated parameter.
        Parameters
        ----------
        k : int
            parameter number.

        """
        return "%s_team_%d" % ( self.PARNAMES[k % self.complexity], k / self.complexity )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        """
        return units.Unit ( 1 )





import numpy as numpy

from .Engine import Engine
from .Engine import DummyPlotter

__author__ = "Do Kester"
__year__ = 2024
__license__ = "GPL3"
__version__ = "3.2.1"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017 - 2024 Do Kester

class RandomEngine( Engine ):
    """
    RandomEngine.

    It generates a random trial point from the available unit box.
    If it is OK (> lowLhood) it is kept. 
    The execute method returns immediately as the new point is random and 
    completely independent of the fiducial point.

    If is is not OK, the trial point is shrunk by a random amount toward 
    the fiducial point, until it is accepted.
    Now the new point is not independent of the fiducial point. We start 
    the execution again starting from the new point. 

    The restart is repeated a few times after which he new point is deemed
    sufficiently independent. 

    Attributes from Engine
    ----------------------
    walkers, errdis, maxtrials, slow, rng, report, phantoms, verbose

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, **kwargs ):
        """
        Constructor.
        Parameters
        ----------
        walkers : WalkerList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : RandomEngine
            engine to be copied
        kwargs : for Engine
            "phantoms", "slow", "seed", "verbose"
        """
        super( ).__init__( walkers, errdis, copy=copy, **kwargs )
        self.plotter = DummyPlotter()

    def copy( self ):
        """ Return copy of this.  """
        return RandomEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "RandomEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False, iteration=0 ):
        """
        Execute the engine by a random selection of the parameters.

        Parameters
        ----------
        kw : int
            index of Walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            set walker in place or append
        iteration : int
            iteration number

        Returns
        -------
        int : the number of successfull moves

        """
        self.reportCall()

        walker = self.walkers[kw].copy()
        problem = walker.problem
        fi = walker.fitIndex
        nf = len( fi )
        param = walker.allpars
        
        ## Make a unit box to find a new walker in
        nap = len( param )
        um, ux = self.getUnitMinmax( problem, lowLhood )

        mlen = len( self.walkers )
        dr = 10 * ( ux - um ) / mlen
        um -= dr
        um = numpy.where( um < 0, 0.0, um )
        ux += dr
        ux = numpy.where( ux > 1, 1.0, ux )

        self.plotter.start( param=walker.allpars )

        t = 0
        for tt in range( self.nstep ) :
            ptry = param.copy()
            uval = self.rng.uniform( um, ux, nap )
            ptry[fi] = self.unit2Domain( problem, uval, kpar=fi )

            Ltry = self.errdis.logLikelihood( problem, ptry )

            if Ltry >= lowLhood :       ## Lucky, in 1 step a truely random point
                self.plotter.move( param, ptry, col=0, sym=2 )
                self.reportSuccess()
                update = len( self.walkers ) if append else kw
                self.setWalker( update, problem, ptry, Ltry, fitIndex=fi )

                self.plotter.stop( param=ptry, name="RandomEngine" )
                return nf                  # nr of successfull params moves

            self.plotter.move( param, ptry, col=5, sym=4 )
            """
            ## not lucky; shrink onto the fiducial point
            utry = self.domain2Unit( problem, param )
            kk = 0
            while True :
                kk += 1
                dr = self.rng.uniform( 0.0, 1.0, 1 )
                uval = ( 1 - dr ) * utry + dr * uval
                ptry[fi] = self.unit2Domain( problem, uval, kpar=fi )

                Ltry = self.errdis.logLikelihood( problem, ptry )
                if Ltry >= lowLhood:
                    self.reportSuccess( )
                    self.plotter.move( param, ptry, col=0, sym=4 )
                    update = len( self.walkers ) if append else kw
                    self.setWalker( update, problem, ptry, Ltry, fitIndex=fi )
                    param = ptry
                    t += 1
                    break
                elif kk < self.maxtrials :
                    self.reportReject( )
                    self.plotter.move( param, ptry, col=5, sym=4 )
                else :
                    self.reportFailed()
                    break
            """
        if t == 0 :
            self.reportFailed()
            return 0

        self.plotter.stop( param=ptry, name="RandomEngine" )

        return nf                        # nr of successfull params moves





import numpy as numpy
from . import Tools

from .LevenbergMarquardtFitter import LevenbergMarquardtFitter
from .Walker import Walker
from .PhantomCollection import PhantomCollection
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2024
__license__ = "GPL3"
__version__ = "3.2.1"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2024 Do Kester

class Engine( object ):
    """
    Engine defines common properties of all Engines.

    An Engine moves around a walker in a random way such that its likelood
    remain above the low-likelihood-limit.

    Attributes
    ----------
    walkers : WalkerList
        list of walkers to be diffused
    errdis : ErrorDistribution
        error distribution to be used
    slow : int
        If slow > 0, run this engine every slow-th iteration.
    phantoms : PhantomCollection
        Collection of valid walker positions collected during engine execution
    maxtrials : int
        maximum number of trials for various operations
    rng : numpy.random.RandomState
        random number generator
    verbose : int
        if verbose > 4 report about the engines. (mostly for debugging)

    report : list of int (read only)
        reports number of succes, accepted, rejected, failed calls. Plus the total.
    unitRange : array_like (read only)
        present max size of the parameter cloud (in unitspace: [0,1])
    unitMin : array_like (read only)
        present minimum values of the parameter cloud (in unitspace: [0,1])

    Author       Do Kester.

    """
    NSTEP  = 5              #  number of steps
    MAXTRIALS = 5           # maximum number of trials

    SUCCESS = 0             #  succesfull move
    REJECT = 1              #  rejected move
    FAILED = 2              #  failed to move
    BEST   = 3              #  better than all
    NCALLS = 4              #  number of calls

    #  *********CONSTRUCTORS***************************************************

    def __init__( self, walkers, errdis, slow=None, phantoms=None, copy=None, 
                    seed=4213, verbose=0 ):
        """
        Constructor.

        Only one PhantomCollection should be present for all Engines.

        Parameters
        ----------
        walkers : list of Walker
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        slow : None or int > 0
            Run this engine every slow-th iteration. None for all.
        phantoms : None or PhantomCollection
            Container for all valid walkers, that have been tried. But were not kept.
            To calculate the spread of the parameters vs likelihood.
        seed : int
            for random number generator
        verbose : int
            report about the engines when verbose > 4
        copy : Engine
            engine to be copied 
        """
        self.walkers = walkers
        self.errdis = errdis
        self.report = [0]*5

        self.checkBest = self.noBoost
 
        if copy is None :
            self.maxtrials = self.MAXTRIALS
            self.nstep = self.NSTEP
            self.rng = numpy.random.RandomState( seed )
            self.verbose = verbose
            if slow is not None : self.slow = slow
            if not phantoms is None :
                self.phantoms = phantoms
        else :
            self.maxtrials = copy.maxtrials
            self.nstep = copy.nstep
            self.rng = copy.rng
            self.verbose   = copy.verbose
            if hasattr( copy, "slow" ) : self.slow = copy.slow
            self.phantoms = copy.phantoms

    def copy( self ):
        """ Return a copy of this engine.  """
        return Engine( self.walkers, self.errdis, copy=self )

    def bestBoost( self, problem, myFitter=None ) :
        """
        When a logL is found better that all the rest, try to update
        it using a fitter.

        Parameters
        problem : Problem
            the problem at hand
        myFitter : None or Fitter
            None fetches LevenbergMarquardtFitter
            a (non-linear) fitter
        """
        self.checkBest = self.doBoost
        if myFitter is None :
            myFitter = LevenbergMarquardtFitter
        self.fitter = myFitter( problem.xdata, problem.model )



    #  *********SET & GET***************************************************
    def setWalker( self, kw, problem, allpars, logL, walker=None, fitIndex=None ) :
        """
        Update the walker with problem, allpars, LogL and logW.

        Parameters
        ----------
        walker : Sample
            sample to be updated

        kw : int
            index in walkerlist, of the walker to be replaced
        problem : Problem
            the problem in the walker
        allpars : array_like
            list of all parameters
        logL : float
            log Likelihood
        walker : Walker or None
            Copy this walker or create new one
        fitIndex : array_like
            (new) fitIndex
        """
        if walker is None :
            id = 0 if kw >= len( self.walkers ) else self.walkers[kw].id
            wlkr = Walker( id, problem, allpars, fitIndex )
        else :
            wlkr = walker.copy()
            wlkr.allpars = allpars    # Tools.toArray( allpars, ndim=1, dtype=float )

        wlkr.logL = logL
        self.checkBest( wlkr )

        self.walkers.setWalker( wlkr, kw )

        self.phantoms.storeItems( wlkr )

    def noBoost( self, walker ) :
        pass

    def doBoost( self, walker ) :
        """
        Check if walker is best in phantoms and try to optimize.

        Parameters
        ----------
        walker : Walker
            new walker to be checked
        """
        wlist = self.phantoms.getList( walker )

        if wlist is None or wlist[-1].logL >= walker.logL :
            return

#        print( "Best   ", fmt( walker.logL ) )
#        print( fmt( walker.allpars, max=None ) )

        problem = walker.problem
        nh = self.errdis.nphypar
        par0 = walker.allpars[:-nh]
        try :
            self.fitter.model = problem.model
            pars = self.fitter.fit( problem.ydata, par0=par0 )
        except :
            return

        if nh == 0 :
            ptry = pars
        elif walker.fitIndex[-1] == -1 :
            ptry = numpy.append( pars, self.fitter.getScale() )
        else :
            ptry = numpy.append( pars, walker.allpars[-1] )

        Ltry = self.errdis.logLikelihood( problem, ptry )

        if Ltry > walker.logL :
            walker.allpars = ptry
            walker.logL = Ltry

#        print( "Better ", fmt( walker.logL ) )
#        print( fmt( walker.allpars, max=None ) )

        self.reportBest()

######## domain <> unit ###########################################

    def domain2Unit( self, problem, dval, kpar=None ) :
        """
        Return value in [0,1] for the selected parameter.

        Parameters
        ----------
        problem : Problem
            the problem involved
        dval : float
            domain value for the selected parameter
        kpar : None or array_like
            selected parameter index, where kp is index in [parameters, hyperparams]
            None means all
        """
        np = problem.npars
        if kpar is None :
            kpar = self.makeIndex( np, dval )

        elif Tools.isInstance( kpar, int ) :
            return ( problem.domain2Unit( dval, kpar ) if kpar >= 0 else
                     self.errdis.domain2Unit( dval, kpar ) )

        uval = numpy.ndarray( len( kpar ), dtype=float )

        for i,kp in enumerate( kpar ) :
            if kp >= 0 :
                uval[i] = problem.domain2Unit( dval[i], kp )
            else :
                uval[i] = self.errdis.domain2Unit( dval[i], kp )
        return uval

    def unit2Domain( self, problem, uval, kpar=None ) :
        """
        Return domain value for the selected parameter.

        Parameters
        ----------
        problem : Problem
            the problem involved
        uval : array_like
            unit value for the selected parameter
        kpar : None or array_like
            selected parameter indices, where kp is index in [parameters, hyperparams]
            None means all.
        """
        np = problem.npars

        if kpar is None :
            kpar = self.makeIndex( np, uval )
        elif Tools.isInstance( kpar, int ) :
            return ( problem.unit2Domain( uval, kpar ) if kpar >= 0 else
                     self.errdis.unit2Domain( uval, kpar ) )

        dval = numpy.ndarray( len( kpar ), dtype=float )
        for i,kp in enumerate( kpar ) :
            if kp >= 0 :
                dval[i] = problem.unit2Domain( uval[i], kp )
            else :
                dval[i] = self.errdis.unit2Domain( uval[i], kp )
        return dval

    def startJourney( self, unitStart ) :
        """
        Calculate the starting position and reset

        Parameters
        ----------
        unitStart : array_like
            start position in npars-dimensions in unit space
        """
        self.journey = 0
        self.jstart = numpy.sqrt( numpy.sum( unitStart * unitStart ) )

    def calcJourney( self, unitDistance ) :
        """
        Calculate the distance travelled since reset

        Parameters
        ----------
        unitDistance : array_like
            step size in npars-dimensions in unit space
        """
        self.journey += numpy.sqrt( numpy.sum( unitDistance * unitDistance ) )

    def reportJourney( self ) :
        try :
            return ( self.jstart, self.journey )
        except :
            return ( 0,0 )

    def makeIndex( self, np, val ) :
        kpar = [k for k in range( np )]
        nh = len( val ) - np
        kpar += [-k for k in range( nh, 0, -1 )]
        return kpar

    def reportCall( self ):
        """ Store a call to engine  """
        self.report[self.NCALLS] += 1

    def reportSuccess( self ):
        """
        Add 1 to the number of succesfull steps: logL < lowLhood.
        """
        self.report[self.SUCCESS] += 1

    def reportReject( self ):
        """
        Add 1 to the number of rejected steps: logL > lowLhood.
        """
        self.report[self.REJECT] += 1

    def reportFailed( self ):
        """
        Add 1 to the number of failed steps: could not construct a step.
        """
        self.report[self.FAILED] += 1

    def reportBest( self ):
        """
        Add 1 to the number of best likelihoods found upto now.
        """
        self.report[self.BEST] += 1

    def printReport( self, best=False ) :
        if best :
            print( " %10d %10d %10d %10d %10d" % (self.report[0], self.report[1],
                              self.report[2], self.report[3], self.report[4] ) )
        else :
            print( " %10d %10d %10d %10d" % (self.report[0], self.report[1],
                              self.report[2], self.report[4] ) )

    def successRate( self ) :
        """
        Return percentage of success.
        """
        if not hasattr( self, "save" ) :
            self.save = [0, 0]

        srate = ( 100 * ( self.report[0] - self.save[0] ) / 
                  ( self.report[0] + self.report[1] + self.report[2] - self.save[1] ) )
        self.save = [self.report[0], ( self.report[0] + self.report[1] + self.report[2] )]
        return srate


    def getUnitMinmax( self, problem, lowLhood ) :
        """
        Calculate unit minimum and maximum from the Phantoms

        Parameters
        ----------
        problem : Problem
            To extract the unit range for
        lowLhood : float
            low likelihood boundary
        """
        npars = problem.npars
        pamin, pamax = self.phantoms.getParamMinmax( lowLhood, np=problem.npars )

        npars += self.errdis.nphypar
        umax = ( numpy.ones( npars, dtype=float )  if pamax is None
                else self.domain2Unit( problem, pamax ) )
        umin = ( numpy.zeros( npars, dtype=float )  if pamin is None
                else self.domain2Unit( problem, pamin ) )
        umin = numpy.fmin( umin, umax )

        return ( umin, umax )

    def getUnitRange( self, problem, lowLhood ) :
        """
        Calculate unit range and minimum from PhantomCollection

        Parameters
        ----------
        problem : Problem
            To extract the unit range for
        lowLhood : float
            low likelihood boundary
        """
        umin, umax = self.getUnitMinmax( problem, lowLhood )
 
        uran = numpy.abs( umax - umin )
                
        return ( uran, umin )

    def __str__( self ) :
        return str( "Engine" )

    def execute( self, kw, lowLhood ):
        """
        Execute the engine for difusing the parameters

        Parameters
        ----------
        kw : walker-id
            walker to diffuse
        lowLhood : float
            low limit on the loglikelihood

        Returns
        -------
        int : number of succesfull moves

        """
        pass


class DummyPlotter( object ) :

    def __init__( self, iter=1 ) :
        self.iter = iter

    def start( self, param=None, ulim=None ):
        """ start the plot. """
        pass

    def point( self, param, col=None, sym=0 ):
        """
        Place a point at position param using color col and symbol sym.
        """
        pass

    def move( self, param, ptry, col=None, sym=None ):
        """
        Move parameters at position param to ptry using color col.
        """
        pass

    def stop( self, param=None, name=None ):
        """ Stop (show) the plot. """
        pass




import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt
from .Walker import Walker

__author__ = "Do Kester"
__year__ = 2024
__license__ = "GPL3"
__version__ = "3.2.1"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2008 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2024 Do Kester


class WalkerList( list ):
    """
    WalkerList is a list of Walker.

    It is the working ensemble of NestedSampler.


    Attributes
    ----------
    logZ : float (read-only)
        Natural log of evidence
    info : float (read-only)
        The information H. The compression factor ( the ratio of the prior space
        available to the model parameters over the posterior space ) is equal to the exp( H ).
    iteration : int
        Present iteration number.

    Author       Do Kester

    """
    def __init__( self, problem=None, ensemble=0, allpars=None, fitIndex=None, 
                  walker=None,  walkerlist=None ):
        """
        Constructor.

        To be valid it needs either problem/allpars/fitindex or walker or walkerlist

        Parameters
        ----------
        problem : Problem or None
            to construct a walker to be added.
        ensemble : int
            number of walkers
        allpars : array_like
            parameters of the problem
        fitIndex : array of int
            list of parameters to be fitted.
        walker : Walker or None
            walker to be added.
        walkerlist : Walkerlist or None
            walkerlist to be incorporated.
        """
        if not walkerlist is None :
            super( WalkerList, self ).__init__( walkerlist )
            self._count = len( walkerlist )
            return

        super( WalkerList, self ).__init__( )
        self._count = 0

        if problem is not None :
            walker = Walker( 0, problem, allpars, fitIndex )

        self.addWalkers( walker, ensemble )


    def addWalkers( self, walker, ensemble ):
        for i in range( ensemble ) :
            wlkr = walker.copy()
            if wlkr.problem.model and wlkr.problem.model.isDynamic() :
                wlkr.fitIndex = walker.fitIndex.copy()
            wlkr.id = self._count
            self.append( wlkr )
            self._count += 1

    # ===========================================================================
    def setWalker( self, walker, index ):
        """
        replace/append a Walker to this list

        Parameters
        ----------
        walker : Walker
            the list to take to copy from
        index : int
            the index at which to set
        """
#        print( "setWalker  ", id( self ), len( self ), walker.id, index )

        if index < len( self ) :
            self[index] = walker

        else :
            walker.id = self._count
            self._count += 1
            self.append( walker )

    def copy( self, src, des, wlist=None, start=0 ):
        """
        Copy one item of the list onto another.

        Parameters
        ----------
        src : int
            the source item
        des : int
            the destination item
        wlist : WalkerList or None
            Copy from this WalkerList (None == self)
        start : int
            iteration where this walker was created
        """
        if wlist is None :
            wlist = self

        id = self[des].id
        self[des] = wlist[src].copy()
        setatt( self[des], "id", id )
        setatt( self[des], "parent", src )
        setatt( self[des], "start", start )
        setatt( self[des], "step", 0 )


    def logPlus( self, x, y ):
        """
        Return the log of sum.
        """
        return numpy.logaddexp( x, y )

    def firstIndex( self, lowL ) :
        """
        Return  index of the first walker with walker.logL > lowL, 
                None if list is empty
                len  if no item applies 

        Parameters
        ----------
        lowL : float
            low Likelihood
        """
        return next( ( k for k,w in enumerate( self ) if w.logL > lowL ), 
                    None if len( self ) == 0 else len( self)  )


    def insertWalker( self, walker ):
        """
        Insert walker to this list keeping it sorted in logL

        Parameters
        ----------
        walker : Walker
            the list to take to copy from
        """
        klow = self.firstIndex( walker.logL )
        if klow is None :
            self.setWalker( walker, 0 )
            return self

        cnt = self._count

        wl = WalkerList( walkerlist=[walker] )

        if len( self[klow:] ) > 0 :
            wl = wl + WalkerList( walkerlist=self[klow:] )

        if len( self[:klow] ) > 0 :
            self = WalkerList( walkerlist=self[:klow] ) + wl
        else :
            self = wl

        self = WalkerList( walkerlist=self )
        self._count = cnt + 1

        return self

    def cropOnLow( self, lowL ) :
        """
        Return WalkerList with all LogL > lowL

        Precondition: self is ordered on logL

        Parameters
        ----------
        lowL : float
            low Likelihood
        """
        cnt = self._count
        self = WalkerList( walkerlist=self[self.firstIndex( lowL ):] )
        self._count = cnt
        return self


    def getLogL( self, walker=None ) :
        """
        Return the logL of the/all walker

        Parameters
        ----------
        walker : None or Walker
            None return value for all walkers
            get the logL from
        """
        if walker is None :
            return self.getLogLikelihoodEvolution()
        return walker.logL


    def allPars( self ):
        """
        Return a 2d array of all parameters.

        In case of dynamic models the number of parameters may vary.
        They are zero-padded. Use `getNumberOfParametersEvolution`
        to get the actual number.

        Parameters
        ----------
        kpar : int or tuple of ints
            the parameter to be selected. Default: all

        """
        nap = max( [len( w.allpars ) for w in self] )

        pe = numpy.zeros( ( len( self ), nap ), dtype=float )
        for k, walker in enumerate( self ) :
            pe[k,:len( walker.allpars )] = walker.allpars

        return pe

    def getParameterEvolution( self, kpar=None ):
        """
        Return the evolution of one or all parameters.

        In case of dynamic models the number of parameters may vary.
        They are zero-padded. Use `getNumberOfParametersEvolution`
        to get the actual number.

        Parameters
        ----------
        kpar : int or tuple of ints
            the parameter to be selected. Default: all

        """
        pe = []
        for walker in self :
            pe += [walker.parameters]
        if kpar is None :
            return numpy.asarray( pe )
        else :
            return numpy.asarray( pe )[:,kpar]

    def getScaleEvolution( self ):
        """ Return the evolution of the scale.  """
        pe = [walker.hypars for walker in self]
        return numpy.asarray( pe )

    def getLogLikelihoodEvolution( self ):
        """ Return the evolution of the log( Likelihood ).  """
        pe = [walker.logL for walker in self]
        return numpy.asarray( pe )

    def getLowLogL( self ):
        """
        Return the lowest value of logL in the walkerlist, plus its index.
        """
        low = self[0].logL
        klo = 0
        k = 0
        for walker in self :
            if walker.logL < low :
                klo = k
                low = walker.logL
            k += 1
        return ( low, klo )

    def __str__( self ) :
        return str( "Walkerlist with %3d walkers" % len( self ))





import numpy as numpy
import math

from .Walker import Walker
from .WalkerList import WalkerList
from . import Tools

__author__ = "Do Kester"
__year__ = 2024
__license__ = "GPL3"
__version__ = "3.2.1"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *           2024 Do Kester

class PhantomCollection( object ):
    """
    Helper class for NestedSamplers Engines to collect all trial walkers
    obtained during the NS run. They are kept ordered according to their logL.  
    They are used to find the minimum and maximum values 
    of the parameter settings as function of the likelihood. 

    There are different methods for static models and for dynamic models. 

    For dynamic models only parameter sets of the proper length are searched.
    The kth item in self.logL belongs to the kth list in self.pars.
    If the model had np parameters then self.logL[np][k] pertain to 
    self.pars[np][k,:] which has np items

    For static models there is only one array of self.logL and one 2-d array 
    od self.pars.

    Attributes
    ----------
    phantoms : WalkerList or dict of { int : WalkerList }
        int         number of parameters in the model
        Wlakerlist  list of (phantom) walkers
    paramMin : array_like or None
        minimum values of the parameters at this stage of lowLhood
        None if too few items of this parameter length is present
    paramMax : array_like or None
        maximum values of the parameters at this stage of lowLhood
        None if too few items of this parameter length is present


    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************

    MINPARS = 5

    def __init__( self, dynamic=False ):
        """
        Constructor.

        Parameters
        ----------
        dynamic : bool
            whether it is a dynamic model
        """
        self.minpars = self.MINPARS
        self.ncalls = 0
            
        if dynamic :
            self.phantoms = {}
            self.lowLhood = {}
            self.paramMax = {}
            self.paramMin = {}
            self.getList = self.getDynamicList
            self.getParamMinmax = self.getDynamic
            self.storeItems = self.storeDynamic
            self.length = self.lengthDynamic
        else :
            self.phantoms = WalkerList( )
            self.lowLhood = -math.inf

    def __str__( self ):
        """ Return the name of this object.  """
        return "PhantomCollection"

    ##  Methods for static models

    def length( self, np=0 ) :
        """
        Return length of internal walkerlist

        Parameters
        ----------
        np : int
            number of parameters (in case of dynamic only)
        """
        return len( self.phantoms )

    def getList( self, walker ) :
        """
        Return the applicable WalkerList

        Parameters
        ----------
        walker : Walker
            return list pertaining to this walker (not used here)
        """
        return self.phantoms
 
    def storeItems( self, walker ) :
        """
        Store both items as arrays.

        Parameters
        ----------
        walker : Walker
            to be added to the PhantomCollection
        """
        self.ncalls += 1
        self.phantoms = self.phantoms.insertWalker( walker )

    def calculateParamMinmax( self, lowLhood, np=0 ):
        """
        Calculate the min and max values of the present parameter values.

        Parameters
        ----------
        lowLhood : float
            lower boundary of the log Likelihood
        np : int
            number of parameters (not used in this implementation)

        """
        self.lowLhood = lowLhood
        self.phantoms = self.phantoms.cropOnLow( lowLhood )

        if len( self.phantoms ) == 0 :
            self.paramMin = None
            self.paramMax = None
        else :
            pars = self.phantoms.allPars()
            self.paramMin = numpy.nanmin( pars, axis=0 )
            self.paramMax = numpy.nanmax( pars, axis=0 )


    def getParamMinmax( self, lowLhood, np=0 ):
        """
        Obtain the min and max values of the present parameter values.

        Parameters
        ----------
        lowLhood : float
            lower boundary of the log Likelihood
        np : int
            number of parameters (not used in this implementation)

        """
        if lowLhood > self.lowLhood :
            self.calculateParamMinmax( lowLhood, np=np )

        return ( self.paramMin, self.paramMax )

    ##  Dynamic methods

    def lengthDynamic( self, np=None ) :
        """
        Return length of internal walkerlist

        Parameters
        ----------
        np : int
            number of parameters (in case of dynamic only)
        """
        if np is None :
            return numpy.sum( [len( ph ) for ph in self.phantoms.values()] )
        return len( self.phantoms[np] ) if np in self.phantoms else 0

    def getDynamicList( self, walker ) :
        """
        Return the applicable WalkerList or None if not present.

        Parameters
        ----------
        walker : Walker
            return list pertaining to this walker
        """
        np = walker.problem.npars
        return self.phantoms[np] if np in self.phantoms else None

    def storeDynamic( self, walker ) :
        """
        Put both items in the dictionaries with npars as key

        Parameters
        ----------
        logL : float
            log Likelihood 
        pars : 1d array
            parameters
        """
        self.ncalls += 1
        np = walker.problem.npars
        if not np in self.phantoms :
            self.phantoms[np] = WalkerList()

        self.phantoms[np] = self.phantoms[np].insertWalker( walker )

    def calculateDynamic( self, lowLhood, np=0 ):
        """
        Calculate the min and max values of the present parameters of length np.

        Parameters
        ----------
        lowLhood : float
            lower boundary of the log Likelihood
        np : int
            number of parameters

        """
        self.lowLhood[np] = lowLhood
        if np in self.phantoms :
            self.phantoms[np] = self.phantoms[np].cropOnLow( lowLhood )
        else :
            self.phantoms[np] = WalkerList()


        if self.length( np ) <= self.minpars :
            self.paramMax[np] = None
            self.paramMin[np] = None
            return

        pars = self.phantoms[np].allPars()
        self.paramMin[np] = numpy.nanmin( pars, axis=0 )
        self.paramMax[np] = numpy.nanmax( pars, axis=0 )

        return

    def getDynamic( self, lowLhood, np=0 ):
        """
        Return the min and max values of the present parameters of length np.

        Parameters
        ----------
        lowLhood : float
            lower boundary of the log Likelihood
        np : int
            number of parameters

        """
        if not np in self.lowLhood or lowLhood > self.lowLhood[np] or self.paramMin[np] is None :
            self.calculateDynamic( lowLhood, np=np )
        return ( self.paramMin[np], self.paramMax[np] )







# run with : python3 -m unittest TestPhantomCollection

import unittest
import numpy as np
import math
import os

from numpy.testing import assert_array_equal as assertAE
from numpy.testing import assert_array_almost_equal as assertAAE

import matplotlib.pyplot as plt
from BayesicFitting import *
from BayesicFitting import formatter as fmt


class Test( unittest.TestCase  ) :

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def test1( self ) :
        print( "====test 1 Static====================" )

        m = PolynomialModel( 3 )
        m.setPrior( 0, UniformPrior( limits=[-5,5] ) )
        problem = ClassicProblem( model=m )

        numpy.random.seed( 12345 )

        phc = PhantomCollection( )
        for k in range( 10 ) :
            pars = numpy.random.randn( 4 ) / ( k + 1 )
            logL = -numpy.sum( numpy.square( pars ) )

            walker = Walker( k, problem, pars, None, logL=logL )
            phc.storeItems( walker )

        for k, ph in enumerate( phc.phantoms ) :
            print( fmt( k ), fmt( ph.logL ), fmt( ph.allpars ) )

        for k in range( 10, 20 ) :
            lowL = numpy.amin( phc.phantoms[0].logL )
            while True :
                pars = numpy.random.randn( 4 ) / ( k + 1 )
                logL = -numpy.sum( numpy.square( pars ) )
                if logL > lowL : break

            walker = Walker( k, problem, pars, None, logL=logL )
            phc.storeItems( walker )

            ur, um = phc.getParamMinmax( lowL )
            print( fmt( k ), fmt( ur ), fmt( um ), phc.length(), fmt( phc.phantoms[0].logL ) )


        for k, ph in enumerate( phc.phantoms ) :
            print( fmt( k ), fmt( phc.phantoms._count ), fmt( ph.logL ), fmt( ph.allpars ) )
        

    def test2( self ) :
        print( "====test 2 Dynamic====================" )

        N = 30
        numpy.random.seed( 12345 )

        phc = PhantomCollection( dynamic=True )
        for k in range( N ) :
            np = numpy.random.randint( 3, high=6 )

            m = PolynomialDynamicModel( np-1 )
            m.setPrior( 0, UniformPrior( limits=[-5,5] ) )
            problem = ClassicProblem( model=m )

            pars = numpy.random.randn( np ) * 30 / ( k + 1 )
            logL = -numpy.sum( numpy.square( pars ) )
            print( fmt( np ), fmt( logL ), fmt( pars ) )
            walker = Walker( k, problem, pars, None, logL=logL )
            phc.storeItems( walker )
            self.assertTrue( np == phc.phantoms[np][0].problem.npars )

        printclass( phc )

        for n in range( 3, 6 ) :
            for k, ph in enumerate( phc.phantoms[n] ) :
                print( fmt( n ), fmt( k ), fmt( ph.logL ), fmt( ph.allpars ) )

#        lowL = min( [numpy.amin( ph.logL ) for ph in phc.phantoms] )
        lowL = min( [numpy.amin( phc.phantoms[k].getLogL() ) for k in phc.phantoms.keys()] )
        low0 = min( [phc.phantoms[k][0].logL for k in phc.phantoms.keys()] )

        print( "low   ", lowL, low0 )
        self.assertTrue( lowL == low0 )

        np = 5
        ur, um = phc.getParamMinmax( lowL, np=np )
        print( "parange  ", fmt( ur ) )
        print( "parmin   ", fmt( um ) )

        for k in range( N, 2*N ) :
            lowL = min( [phc.phantoms[k][0].logL for k in phc.phantoms.keys()] )
            np = numpy.random.randint( 3, high=6 )
            while True :
                pars = numpy.random.randn( np ) * 30 / ( k + 1 )
                logL = -numpy.sum( numpy.square( pars ) )
                if logL > lowL : break
            walker = Walker( k, problem, pars, None, logL=logL )
            phc.storeItems( walker )

            ur, um = phc.getParamMinmax( lowL, np=np )

            print( fmt( k ), fmt( np ), fmt( phc.length( np ) ), fmt( lowL ) )
            print( "parange  ", fmt( ur ) )
            print( "parmin   ", fmt( um ) )

            if phc.length( np ) <= phc.minpars :
                self.assertTrue( ur is None )
            else :
                self.assertFalse( um is None )

        for n in range( 3, 6 ) :
            for k, ph in enumerate( phc.phantoms[n] ) :
                print( fmt( n ), fmt( phc.phantoms[n]._count ), fmt( k ), fmt( ph.logL ), fmt( ph.allpars ) )

        s = 0
        for np in phc.phantoms :
            s += phc.length( np )
            print( np, phc.length( np ) )
        print( phc.length() )        
        self.assertTrue( s == phc.length() )

if __name__ == '__main__':
    unittest.main( )




import numpy as numpy
import math
import sys

#from .ErrorDistribution import ErrorDistribution
from .ScaledErrorDistribution import ScaledErrorDistribution
from .LinearModel import LinearModel
from .Fitter import Fitter
from .BaseFitter import BaseFitter
from .CurveFitter import CurveFitter
from .AmoebaFitter import AmoebaFitter
from .LevenbergMarquardtFitter import LevenbergMarquardtFitter
from .Formatter import formatter as fmt
from .Tools import setAttribute as setatt


__author__ = "Do Kester"
__year__ = 2024
__license__ = "GPL3"
__version__ = "3.2.1"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"


#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2019 - 2024 Do Kester


class ModelDistribution( ScaledErrorDistribution ):
    """
    To calculate the probability of a model M from a set of models S,
    given some data D, use Bayes rule:

        P( M|DS ) = P( M|S ) * P( D|MS ) / P( D|S )
        posterior = prior   * likelihood / evidence

    This class calculates the likelihood P( D|MS ).
    On another level where we calculate the probability of the
    parameters p, we see this likelhood appear as evidence P( D|M ).

    Again using Bayes :

        P( p|DM ) = P( p|M ) * P( D|pM ) / P( D|M )

    The evidence here is calculated as the integral over a Gausian
    approximation of the posterior.

    Author       Do Kester.

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, arbiter=None, scale=1.0, limits=None, 
                  copy=None, **kwargs ):
        """
        Default Constructor.

        Parameters
        ----------
        arbiter : None or BaseFitter or str
            to provide the evidence
            None    select fitter automatically
            BaseFiter   Use this fitter
            str     "fitter", "levenberg", "curve", "amoeba", "NestedSampler"

        scale : float
            noise scale
        limits : None or list of 2 floats [low,high]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            when limits are set, the scale is *not* fixed.

        copy : ModelDistribution
            distribution to be copied.

        kwargs : dict
            to be applied to arbiter

        """
        setatt( self, "arbiter", arbiter )
        setatt( self, "kwarbs", kwargs )

        super( ).__init__( limits=limits, scale=scale, copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return ModelDistribution( copy=self, arbiter=self.arbiter )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return True


    #  *********LIKELIHOODS***************************************************
    def logLikelihood_alt( self, problem, allpars ) :
        """
        Return the log( likelihood ) for a Gaussian distribution.

        Alternate calculation

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
            Return optimal parameters of the fit

        """
        mdl = problem.model

        if allpars is None :
            allpars = self.rng.rand( mdl.npars + self.nhypar )
            allpars = problem.unit2Domain( allpars )

        try :
            noiselim = self.hyperpar[0].getLimits()
            fscale = None
        except :
            noiselim = None
            fscale = self.scale

        if noiselim is None and mdl.npars < len( allpars ) :
            allpars[-1] = self.hyperpar[0].hypar

        kwarbs = self.kwarbs

        if self.arbiter is None :
            if isinstance( mdl, LinearModel ) :
                ftr = Fitter( problem.xdata, mdl, fixedScale=fscale, **kwarbs )
            else :
                ftr = LevenbergMarquardtFitter( problem.xdata, mdl, fixedScale=fscale,
                            **kwarbs )
        elif isinstance( self.arbiter, BaseFitter ) :
            ftr = self.arbiter

        elif not isinstance( self.arbiter, str ) :
            raise ValueError( "Cannot interpret %s as string or fitter" % self.arbiter )

        else :
            name = str.lower( self.arbiter )

            if name == "fitter" :
                ftr = Fitter( problem.xdata, mdl, fixedScale=fscale, **kwarbs )
            elif name.startswith( "leven" ) :
                ftr = LevenbergMarquardtFitter( problem.xdata, mdl, fixedScale=fscale, 
                            **kwarbs )
            elif name.startswith( "curve" ) :
                ftr = CurveFitter( problem.xdata, mdl, fixedScale=fscale, **kwarbs )
            elif name.startswith( "amoeba" ) :
                ftr = AmoebaFitter( problem.xdata, mdl, fixedScale=fscale, **kwarbs )

            elif name.startswith( "nested" ) :

                ## Import NestedSampler here to avoid circular imports
                from .NestedSampler import NestedSampler
                ######################################################

                ## we need a static version of the model
                model = mdl.copy( modifiable=False, dynamic=False )

                ## add items to kwarbs if not already present
                if not "ensemble" in kwarbs :
                    kwarbs["ensemble"] = 10
                if not "verbose" in kwarbs :
                    kwarbs["verbose"] = 0
                if not "engines" in kwarbs :
                    kwarbs["engines"] = ["chord", "step"]
                ns = NestedSampler( model=model, xdata=problem.xdata, ydata=problem.ydata,
                        weights=problem.weights, limits=noiselim, **kwarbs )

                try :
                    evidence = ns.sample()
                    allpars[:mdl.npars] = ns.parameters
                    if noiselim is not None :
                        allpars[-1] = ns.scale
                except RuntimeError :
                    evidence = -math.inf
                return evidence

            else :
                raise ValueError( "Cannot interpret %s" % self.arbiter )

#        print( "MD Fitr    ", ftr ) 

        try :
            pars = ftr.fit( problem.ydata )
            evidence = ftr.getLogZ( limits=mdl.getLimits(), noiseLimits=noiselim )

            allpars[:mdl.npars] = pars
            if noiselim is not None :
                allpars[-1] = ftr.scale

#            print( problem.npars, mdl.npars, noiselim, ftr.scale, allpars )

        except numpy.linalg.LinAlgError :
            evidence = - sys.float_info.max
        except ValueError :                         # math domain error
            evidence = - sys.float_info.max


        return evidence

    def logLdata( self, problem, allpars, mockdata=None ) :
        """
        Return the log( likelihood ) for each residual
           
        logL = sum( logLdata )

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem   
        mockdata : array_like
            as calculated by the model
        
        """
        return self.logLikelihood_alt( problem, allpars )


    def __str__( self ) :
        return "ModelDistribution"



import numpy as numpy
import math
from . import Tools
from .Formatter import formatter as fmt

from .OrderEngine import OrderEngine

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.2.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Alpha"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2023 Do Kester

class ShuffleEngine( OrderEngine ):
    """
    The ShuffleEngine tries to shuffle a selection of the parameters in place.

    Input order : [0,1,2,3,4,5,6,7,8,9]

    output order: [0,1,2,5,3,7.4,6,8,9]

    It belongs to the class of generalized travelling salesman problems
    where the parameters of the problem is an ordered list.

    The walker is kept when the logLikelihood > lowLhood

    Attributes from Engine
    ----------------------
    walkers, errdis, maxtrials, nstep, slow, rng, report, phantoms, verbose

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, **kwargs ):
        """
        Constructor.

        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : ShuffleEngine
            to be copied
        kwargs : dict for Engine
            "phantoms", "slow", "seed", "verbose"

        """
        super( ).__init__( walkers, errdis, copy=copy, **kwargs )                  

    def copy( self ):
        """ Return copy of this.  """
        return ShuffleEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "ShuffleEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood ):
        """
        Execute the engine by diffusing the parameters.

        Parameters
        ----------
        kw : int
            id of walker to diffuse
        lowLhood : float
            lower limit in logLikelihood

        Returns
        -------
        int : the number of successfull moves

        """
        walker = self.walkers[kw]
        problem = walker.problem
        np = problem.npars
        param = walker.allpars

        src = self.rng.randint( np )
        t = 2 + min( self.rng.geometric( 0.2 ), np // 2 )

        while t > 1 :
            ent = src + t
            ptry = numpy.roll( param, src )
            ptry[:ent] = self.rng.permutation( ptry[:ent] )


            Ltry = self.errdis.logLikelihood( problem, ptry )

            if self.verbose > 4 :
                print( "Shuffle  ", src, ent, t, fmt( lowLhood ), fmt( Ltry ) )
                print( fmt( param, max=None, format='%3d' ) )
                print( fmt( ptry, max=None, format='%3d' ) )

            if Ltry >= lowLhood:
                self.reportSuccess( )
                self.setWalker( walker.id, problem, ptry, Ltry )

                return t

            t = self.rng.randint( t )
            self.reportFailed( )

        self.reportReject()

        return 0                        # nr of reordered parameters





# run with : python3 -m unittest TestFootbalModel.Test.test1

import unittest
import os
import math
import numpy as numpy
from astropy import units
from astropy import modeling
#from astropy.modeling.models import Gaussian1D
#from astropy.modeling.models import Polynomial1D
import matplotlib.pyplot as plt
import warnings

from numpy.testing import assert_array_almost_equal as assertAAE
from numpy.testing import assert_array_equal as assertAE
from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester


class Test( unittest.TestCase ):
    """
    Test harness for Models

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def test1( self ):
        print( "******Football MODEL 1***********************" )

        m1 = FootballModel( 3, complexity=1  )

        p1 = numpy.asarray( [10, 6.0, 3.0], dtype=float )
        xdata = numpy.asarray( [[0,1], [1,2], [2,0]], dtype=int )

        print( fmt( p1 ) )

        r1 = m1.result( xdata, p1 )
        self.rprint( xdata, r1 )
        print( m1.lastndout, m1.ndout )

        m5 = FootballModel( 3 )

        p5 = numpy.asarray( [10, 0, 1, 1, 1, 6, 0, 1, 1, 1, 3.0, 0, 1, 1, 1], dtype=float )

        print( fmt( p5, max=None )  )

        r5 = m5.result( xdata, p5 )
        self.rprint( xdata, r5 )

        assertAE( r1, r5 )

        k = [1,6,11]
        p5[k] = [0.5, 0.7, 0.9]

        r5 = m5.result( xdata, p5 )
        print( fmt( p5, max=None )  )
        self.rprint( xdata, r5 )

        k = [2,7,12]
        p5[k] = [1.5, 0.7, 1.0]

        r5 = m5.result( xdata, p5 )
        print( fmt( p5, max=None )  )
        self.rprint( xdata, r5 )

    def rprint( self, x, r ) :
        for k in range( x.shape[0] ) :
            print( "%d - %d result %8.3f - %8.3f" % ( x[k,0], x[k,1], r[k,0], r[k,1] ) ) 
        


    def test2( self ):
        print( "******FOOTBALL MODEL 2***********************" )

        N = 10                                          # nr of teams
        M = N // 2                                      # nr of matches
        xdata = numpy.arange( N, dtype=int ).reshape( (5,2) )
        print( xdata )
        mf = numpy.linspace( 0.4, 1.6, M )

        print( "== midfield ====" )

        m3 = FootballModel( N, complexity=3 )
        p3 = numpy.asarray( [10, 0.5, 1] * N, dtype=float )

        p3[2::6] = mf
        print( "p3   ", p3.reshape( (-1,3) ) )
        r3 = m3.result( xdata, p3 )
        rh = r3[:,0]
        ra = r3[:,1]

        print( "home ", fmt( rh ) )
        print( "away ", fmt( ra ) )

        if self.doplot :
            plt.plot( mf, rh, 'k.' )
            plt.plot( mf, ra, 'r.' )

        print( "== home advantage ====" )

        m4 = FootballModel( N, complexity=4 )
        p4 = numpy.asarray( [10, 0.5, 1, 1] * N, dtype=float )

        p4[7::8] = mf
        print( "p4   ", p4.reshape( (-1,4) ) )
        r4 = m4.result( xdata, p4 )
        rh = r4[:,0]
        ra = r4[:,1]

        if self.doplot :
            plt.plot( mf, rh, 'k-' )
            plt.plot( mf, ra, 'r-' )


    def test5( self ):
        print( "******FOOTBALL MODEL 5***********************" )

        N = 10                                          # nr of teams
        M = N // 2                                      # nr of matches
        xdata = numpy.arange( N, dtype=int ).reshape( (5,2) )
        print( xdata )
        mf = numpy.linspace( 0.4, 1.6, M )

        print( "== strategy ====" )

        m5 = FootballModel( N, complexity=5 )
        p5 = numpy.asarray( [10, 0.5, 1, 1, 1] * N, dtype=float )

        p5[4::10] = mf
        print( "p5   \n", p5.reshape( (-1, 5) ) )
        r5 = m5.result( xdata, p5 )
        rh = r5[:,0]
        ra = r5[:,1]

        print( "home ", fmt( rh ) )
        print( "away ", fmt( ra ) )
        if self.doplot :
            plt.plot( mf, rh, 'k-.' )
            plt.plot( mf, ra, 'r-.' )

        p5 = numpy.asarray( [10, 0.5, 1, 1, 1] * N, dtype=float )

        c5 = m5.copy()

        p5[2::10] = 1.2
        p5[9::10] = mf
        print( "p5   \n", p5.reshape( (-1, 5) ) )
        r5 = c5.result( xdata, p5 )
        rh = r5[:,0]
        ra = r5[:,1]

        if self.doplot :
            plt.plot( mf, rh, 'b-.' )
            plt.plot( mf, ra, 'g-.' )
            plt.show()

        for k in range( 0, 5*N, 6 ) :
            print( k, c5.getParameterName( k ), m5.getParameterUnit( k ) )



    def test3( self ):
        print( "******FOOTBALL MODEL test 3***********************" )

        xdata = numpy.asarray( [[0,1], [1,2]], dtype=int )


        numpy.random.seed( 123456 )
        for k in range( 1, 6 ) :
            print( "------- Complexity %d --------------" % k )
            print( "k0   ", xdata[:,0] * k )
            print( "k1   ", xdata[:,1] * k )
            mdl = FootballModel( 3, complexity=k )

            par = numpy.random.rand( 3 * k )
            kk = numpy.asarray( range( 0, 3*k, k ) )
            par[kk]*= 10
            for i in range( 2, k ) :
                par[kk+i] *= 2
            print( k, fmt( par, max=None ) )
            
            res = mdl.result( xdata, par )
            self.rprint( xdata, res )
            print( res )   

            
            part = mdl.partial( xdata, par )
            print( "partial" )
            print( fmt( part, max=None ) )
            nump = mdl.strictNumericPartial( xdata, par )
            print( "numeric" )
            print( fmt( nump, max=None ) )
            assertAAE( part, nump )

    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestUserModel.__class__ )

if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from astropy import units
import math
import warnings
from . import Tools
from .Formatter import formatter as fmt
from .Formatter import fma

from .Engine import Engine
from .Engine import DummyPlotter

import matplotlib.pyplot as plt

__author__ = "Do Kester"
__year__ = 2024
__license__ = "GPL3"
__version__ = "3.2.1"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2024 Do Kester

class GalileanEngine( Engine ):
    """
    Move all parameters in forward steps, with optional mirroring on the edge.

    Move the parameters in a random direction for N iterations; mirror the direction
    on the gradient of the logLikelihood when the parameters enter the zone of logLlow.

    Attributes
    ----------
    size : 0.5
        of the step

    Attributes from Engine
    ----------------------
    walkers, errdis, maxtrials, nstep, slow, rng, report, phantoms, verbose

    Author       Do Kester.

    """

    SIZE = 0.5

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, **kwargs ):
        """
        Default Constructor.

        Parameters
        ----------
        walkers : WalkerList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : GalileanEngine
            to be copied
        kwargs : for Engine
            "phantoms", "slow", "seed", "verbose"

        """
        super( ).__init__( walkers, errdis, copy=copy, **kwargs )

        self.size = self.SIZE

        self.plotter = DummyPlotter( )

    def copy( self ):
        """ Return copy of this.  """
        engine = GalileanEngine( self.walkers, self.errdis, copy=self )
        engine.nstep = self.nstep

        return engine

    def __str__( self ):
        """ Return the name of this engine.  """
        return str( "GalileanEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False, iteration=0 ):
        """
        Execute the engine by diffusing the parameters.

        Parameters
        ----------
        kw : int
            index in walkerlist, of the walker
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            set walker in place of append

        Returns
        -------
        int : the number of successfull moves

        """
        self.reportCall()

        walker = self.walkers[kw].copy()
        problem = walker.problem
        Lhood = walker.logL
        allpars = walker.allpars
        fitIndex = walker.fitIndex

        inside = 0
        Ltry = 0
        size = self.size

        self.plotter.start( param=allpars )
        um = UnitMovements( walker, self, size, lowLhood )

#        nstep = int( self.nstep * ( 1 + self.rng.rand() ) )
        nstep = 2 + int( self.nstep * ( 1 + self.rng.rand() ) )

        maxtrial = self.maxtrials

        ptry = allpars.copy()
        if self.verbose > 4 :
            print( "Galilean   LogL  ", fmt( Lhood ), " LowL  ", fmt( lowLhood), nstep, maxtrial, size )
            print( "alpar ", fma( ptry, linelength=200 ) )
            fip = allpars[fitIndex]
            print( "uap   ", fma( self.domain2Unit( problem, fip, fitIndex ), linelength=200) )
            print( "unitr ", fma( um.upran, linelength=200 ) )
            print( '--------------------' )
        step = 0
        trial = 0

#        um.setParameters( problem, allpars )
#        self.startJourney( um.upar )

        while True:
            trial += 1
            f = -1

            um.setParameters( problem, allpars )        # restart at safe location

            if inside == 0 :                            # safely inside lowLhood area
                ptry[fitIndex] = um.trialStep( 1.0, size )

            elif inside == 1 :                          # first time outside -> mirror

                ptry[fitIndex] = um.trialStep( 0.5, size )    # params at mid point
                Lmid = self.errdis.logLikelihood( problem, ptry )

                f = self.quadinterpol( Lhood, Lmid, Ltry, lowLhood )

                pedge = allpars.copy()
                pedge[fitIndex] = um.trialStep( f, size )      # params at edge  

                if self.verbose > 4 :
                    Ledge = self.errdis.logLikelihood( problem, pedge )
                    print( "Lmid  ", fmt( iteration ), fmt( Lmid ), fmt( f ), fmt( Ltry ),
                            fmt( lowLhood ), fmt( Ledge ), fmt( Lhood ) )

                dLdp = self.errdis.partialLogL( problem, pedge, fitIndex )
                self.plotter.move( allpars, pedge, col=1, sym=4 )
                um.acceptTrial()

                um.mirrorOnLowL( dLdp )

                ptry[fitIndex] = um.trialStep( 1 - f, size )

                self.plotter.move( pedge, ptry, col=2 )
            else:                                       # mirroring failed; do reverse
                size *= 0.7
                um.reverseVelocity( size )
                ptry[fitIndex] = um.trialStep( 1.0, size )

                self.plotter.move( allpars, ptry, col=3, sym=0 )

            Ltry = self.errdis.logLikelihood( problem, ptry )


            if Ltry >= lowLhood:
                self.plotter.move( allpars, ptry, col=0, sym=0 )
#                self.calcJourney( um.upar - um.uptry )

                um.acceptTrial()

                allpars = ptry.copy( )
                Lhood = Ltry
                self.reportSuccess( )
                inside = 0
                step += 1
                trial = 0
                size = self.size

                ## update the walker
                update = len( self.walkers ) if append else kw
                self.setWalker( update, problem, allpars, Lhood, walker=walker,
                                fitIndex=fitIndex )

            else:
                inside += 1
                if inside == 1:
                    self.plotter.move( allpars, ptry, col=5, sym=4 )
                    self.reportReject( )
                else:
                    self.plotter.move( allpars, ptry, col=4 )
                    self.reportFailed( )

            if self.verbose > 4 :
                print( "upar  ", fma( um.upar, linelength=200 ) )
                print( "uvel  ", fma( um.uvel, linelength=200 ) )
                print( "ptry  ", fma( ptry, linelength=200 ) )
                print( "Ltry  ", fmt( Ltry ), inside, step, trial, self.plotter.iter-1, 
                        fmt( size ) )
                print( '--------------------' )


            if not ( step < nstep and trial < maxtrial ):
                break


        # dynamically adjust step size
        if step == 0 :
            self.size = max( 0.9 * self.size, 1e-6 )
            if self.verbose > 4 :
                warnings.warn( "GalileanEngine: no steps found" )
        else :
            self.size = min( self.SIZE, self.size * 1.1 )

        self.plotter.stop( param=allpars, name="GalileanEngine" )

        return step * len( fitIndex )           # nr of successfull steps

    def quadinterpol( self, L0, Lm, L1, lowL ) :
        """
        Quadratic interpolation of points (x,y)
        x = [0.0, 0.5, 1.0]
        y = [L0, Lm, L1]  where L0 > Lm > L1    
        interpolation at y = lowL.

        Returns
        -------
        xvalue : float
            largest of the two inside [0,1]
        """ 
        c = L0
        b = 4 * Lm - L1 - 3 * L0
        a = 2 * L0 - 4 * Lm  + 2 * L1

        if a == 0 :
            return 1 if b == 0 else ( lowL - c ) / b

        det = b * b - 4 * a * ( c - lowL )
        det = 0 if det < 0 else math.sqrt( det )

        x1 = ( - b - det ) / ( 2 * a )
        x2 = ( - b + det ) / ( 2 * a )

        if a < 0 :      # top parabola: return largest
            return x1 if x1 > x2 else x2
        else :          # valley parabala : return smallest
            return x1 if x1 < x2 else x2



class UnitMovements( object ):
    """
    Define all parameter movements in unitspace.



    """
    def __init__( self, walker, engine, size, lowLhood ):
#        print( "GEUM  ", engine.unitRange )
        self.problem = walker.problem
        self.fitIndex = walker.fitIndex

        self.np = len( self.fitIndex )
        self.engine = engine
        self.setParameters( self.problem, walker.allpars )

        uran, umin = engine.getUnitRange( self.problem, lowLhood )
        self.upran = uran[self.fitIndex]

        self.upran = numpy.amax( self.upran )
        self.uvel = self.getVelocity( size )

    def uniform( self ) :
        """
        Return random step between [0.5,0.5]
        """
        return self.engine.rng.rand( self.np ) - 0.5

    def setParameters( self, problem, allpars ):
        """
        Set unit values for the applicable parameters
        """
        fip = allpars[self.fitIndex]
        self.upar = self.engine.domain2Unit( problem, fip, kpar=self.fitIndex )

    def trialStep( self, f, size ):
        """
        Return a new trial parameter set.
        """
        uv = self.uvel
        up = self.upar + uv * f

        # check if outside [0,1]
        up, uv = numpy.where( up <= 0, ( -up, -uv ), ( up, uv ) )
        self.uptry, self.uvtry = numpy.where( up >= 1, ( 2 - up, -uv ), ( up, uv ) )

        # When on edge, start a new direction with the same signs
        if not all( self.uvtry == self.uvel ) :
            uv = self.getVelocity( size ) 
            self.uvtry = numpy.copysign( uv, self.uvtry )

        return self.engine.unit2Domain( self.problem, self.uptry, kpar=self.fitIndex )

    def acceptTrial( self ) :
        self.upar = self.uptry
        self.uvel = self.uvtry

    def mirrorOnLowL( self, dLdp ):
        """
        Mirror the velocity on the low likelihood border.
        """
        inprod = numpy.sum( dLdp * self.uvel )
        sumsq  = numpy.sum( dLdp * dLdp )
        if sumsq == 0 : return

        self.uvel -= 2 * dLdp * inprod / sumsq


    def getVelocity( self, size ):
        """
        Return a new (random) unit velocity.
        """
        return self.uniform() * self.upran * size

    def reverseVelocity( self, size ):
        """
        Reverse the unit velocity when mirroring fails.
        Add small perturbation.
        """
        ff = 0.1                                # 10 percent
        uvpert = self.getVelocity( ff * size )  # unit velocity perturbance
        self.uvel *= ( ff - 1 )                 # reverse uvel and take 90 %
        self.uvel += uvpert                     # add perturbance

        """
        TBC I dont think this is any good.

        uv = self.uvel                 # keep original velocity
        self.uvel = self.getVelocity( size )        # get a new one to perturb
        nm = len( self.engine.walkers )
        self.uvel = size * ( self.np * self.uvel - nm * upv ) / ( nm + self.np )

        """

